# -*- coding: utf-8 -*-
"""
@author: jsch
Version 0.1.a
Demonstration of the radial coordinate inversion of the Webservices inversion routine
as applied to the 5/5 island chain of the 'Iota Scan' configurations, Low Iota and High Iota.
Also see: Tamara Andreeva et al 2021 Nucl. Fusion in press https://doi.org/10.1088/1741-4326/ac3f1b

To run the examples, uncomment the appropriate lines at the end of this source file

Coordinates can be transformed to acquire maps at other field periods by adding or 
subtracting factors of N*(2*pi)/5, where N is an integer.

"""

# Import libraries
import os
import numpy as np
import time
import matplotlib.pyplot as plt
from osa import Client
import h5py

# define field line client
fieldline_client = Client("http://esb:8280/services/FieldLineProxy?wsdl")

pos = fieldline_client.types.Points3D()
    
# Define and instantiate the vmec service client.  'v9' is the newest version 
vmec_client = Client('http://esb:8280/services/vmec_v9?wsdl')


inv_tol = 1e-6

# Define the configurations in the webesrvices vmec services database
vmec_main_configs = {}

#

# 5/5 Island configurations
vmec_configs = {}
vmec_configs['eim'] =    'jsch_eim_edge_island_v1_fixed_v1'  
vmec_configs['eim_fix'] = 'jsch_eim_edge_island_v1_fixed_v1'       


# These are the 2nd versions. The islands are much larger, Axis accuracy ~ a few mm?
vmec_configs['ejm252'] = 'jsch_ejm252_edge_6252mm_v0a'       
vmec_configs['ejm004'] = 'jsch_ejm004_edge_6268mm_v0a'       
vmec_configs['flm001'] = 'jsch_flm001_edge_6215mm_v0a'  
vmec_configs['fmm'] =    'jsch_fmm_edge_6196mm_v0a'    
vmec_configs['fmm001'] = 'jsch_fmm001_edge_6201mm_v0a'  
vmec_configs['fmm002'] = 'jsch_fmm002_edge_6206mm_v0a'  
vmec_configs['fmm003'] = 'jsch_fmm003_edge_6210mm_v0a'  
vmec_configs['fnm'] =    'jsch_fnm_edge_6181mm_v0a'
vmec_configs['fnm001'] = 'jsch_fnm001_edge_6187mm_v0a'  
vmec_configs['fnm002'] = 'jsch_fnm002_edge_6192mm_v0a'  
vmec_configs['fom003'] = 'jsch_fom003_edge_6176mm_v0a'
vmec_configs['fqm001'] = 'jsch_fqm001_edge_6128mm_v0a'



# Get the 'keys' for all of the vmec configurations that are defined.
all_cases = vmec_configs.keys()

# Now, define a matching set of configurations for line following
# (coil model and coil currents)
linefollow_configs = {}

for this_case in all_cases:    
    linefollow_configs[this_case] = fieldline_client.types.MagneticConfig()
    # All of the cases will use the same set of coils.  See the coils database for details
    linefollow_configs[this_case].coilsIds = list(np.arange(160, 230))
    linefollow_configs[this_case].coilsIds.extend([230, 231, 232, 233, 234, 235, 236, 237, 238, 239]) 

# Now, define the coil currents.
# 'eim', eim252 do not have iota correction.  All of the rest DO HAVE IOTA CORRECTION

this_case = 'eim'
linefollow_configs[this_case].coilsIdsCurrents = [12989.*108, 12989.*108, 12989.*108, 12989.*108, 12989.*108]*10
linefollow_configs[this_case].coilsIdsCurrents.extend([0.*36, 0.*36]*10)
linefollow_configs[this_case].coilsIdsCurrents.extend([0*8]*10)

this_case = 'eim_fix'
linefollow_configs[this_case].coilsIdsCurrents = [12989.*108, 12989.*108, 12989.*108, 12989.*108, 12989.*108]*10
linefollow_configs[this_case].coilsIdsCurrents.extend([0.*36, 0.*36]*10)
linefollow_configs[this_case].coilsIdsCurrents.extend([0*8]*10)


this_case = 'ejm252'
linefollow_configs[this_case].coilsIdsCurrents = [13114.*108]*50
linefollow_configs[this_case].coilsIdsCurrents.extend([-574.*36, -574.*36]*10)
linefollow_configs[this_case].coilsIdsCurrents.extend([0*8, 0*8]*5)

this_case = 'ejm004'
linefollow_configs[this_case].coilsIdsCurrents = [13016.*108]*50
linefollow_configs[this_case].coilsIdsCurrents.extend([242.*36, 242.*36]*10)
linefollow_configs[this_case].coilsIdsCurrents.extend([0*8, 0*8]*5)


this_case = 'flm001'
linefollow_configs[this_case].coilsIdsCurrents = [13361.*108]*50
linefollow_configs[this_case].coilsIdsCurrents.extend([-2641.*36, -2641.*36]*10)
linefollow_configs[this_case].coilsIdsCurrents.extend([0*8, 0*8]*5)

this_case = 'fmm'
linefollow_configs[this_case].coilsIdsCurrents = [13485.*108]*50
linefollow_configs[this_case].coilsIdsCurrents.extend([-3674.*36, -3674.*36]*10)
linefollow_configs[this_case].coilsIdsCurrents.extend([0*8, 0*8]*5)

this_case = 'fmm001'
linefollow_configs[this_case].coilsIdsCurrents = [13454.*108]*50
linefollow_configs[this_case].coilsIdsCurrents.extend([-3416.*36, -3416.*36]*10)
linefollow_configs[this_case].coilsIdsCurrents.extend([0*8]*10)
             
this_case = 'fmm002'
linefollow_configs[this_case].coilsIdsCurrents = [13423.*108]*50
linefollow_configs[this_case].coilsIdsCurrents.extend([-3157.*36, -3157.*36]*10)
linefollow_configs[this_case].coilsIdsCurrents.extend([0*8, 0*8]*5)

this_case = 'fmm003'
linefollow_configs[this_case].coilsIdsCurrents = [13392.*108]*50
linefollow_configs[this_case].coilsIdsCurrents.extend([-2899.*36, -2899.*36]*10)
linefollow_configs[this_case].coilsIdsCurrents.extend([0*8, 0*8]*5)

this_case = 'fnm'
linefollow_configs[this_case].coilsIdsCurrents = [13577.*108]*50
linefollow_configs[this_case].coilsIdsCurrents.extend([-4449.*36, -4449.*36]*10)
linefollow_configs[this_case].coilsIdsCurrents.extend([0*8, 0*8]*5)

this_case = 'fnm001'
linefollow_configs[this_case].coilsIdsCurrents = [13546.*108]*50
linefollow_configs[this_case].coilsIdsCurrents.extend([-4191.*36, -4191.*36]*10)
linefollow_configs[this_case].coilsIdsCurrents.extend([0*8]*10)

this_case = 'fnm002'
linefollow_configs[this_case].coilsIdsCurrents = [13515.*108]*50
linefollow_configs[this_case].coilsIdsCurrents.extend([-3932.*36, -3932.*36]*10)
linefollow_configs[this_case].coilsIdsCurrents.extend([0*8]*10)
    
this_case = 'fom003'
linefollow_configs[this_case].coilsIdsCurrents = [13608.*108]*50
linefollow_configs[this_case].coilsIdsCurrents.extend([-4707.*36, -4707.*36]*10)
linefollow_configs[this_case].coilsIdsCurrents.extend([0*8, 0*8]*5)

this_case = 'fqm001'
linefollow_configs[this_case].coilsIdsCurrents = [13883.*108]*50
linefollow_configs[this_case].coilsIdsCurrents.extend([-7032.*36, -7032.*36]*10)
linefollow_configs[this_case].coilsIdsCurrents.extend([0*8, 0*8]*5)

# END OF magnetic configuration definitions


# Define toroidal cuts and the 'regions of interest' for each toroidal cut

# Step 1
# Grid dimensions and # of toroidal angles

# each toroidal angle will have a grid that is num_pts_grid x num_pts_grid in size
#num_pts_grid = 701  # 201 is decent, 501 is really dense
#num_pts_grid = 101  # 201 is decent, 501 is really dense
num_pts_grid = 201  # 201 is decent, 501 is really dense
#num_pts_grid = 81  # 201 is decent, 501 is really dense
#num_pts_grid = 15

num_poincare_launch_points = 12 # 21 51 is decent for plots 
#num_poincare_launch_points =21 # 21 51 is decent for plots 
#num_poincare_points_per_plane =  1200 # 20 is fast. 50 is good for some detail, 200 is really dense
num_poincare_points_per_plane =  50 # 20 is fast. 50 is good for some detail, 200 is really dense
#num_poincare_points_per_plane =  5 # 20 is fast. 50 is good for some detail, 200 is really dense

# 11 toroidal angles generates 1 for each 1/2 field period + plus one extra
num_toroidal_angles = 10
tor_angles_plus_one_extra = np.linspace(0, 2*np.pi, num_toroidal_angles+1)
tor_angles = tor_angles_plus_one_extra[:-1] # Trim off the last element

# Define min and max limits of R and Z values of inversion maps
# There should be some symmery to these numbers
Rmin = [5.9,   5.3,   5.35,    4.6,   5.3,   4.4,  5.3,  4.6,  5.35,  5.3 ]
Rmax = [6.32, 6.25,    6.1,    5.4,   5.85,  4.91, 5.85, 5.4,  6.1,   6.25]
Zmin = [-.6,    0 ,    .35,   0.15,    0.,  -.35, -.95, -.75, -1.15, -.45]
Zmax = [0.6,  .45 ,   1.15,   0.75,  0.95 ,  .35,   0., -.15, -.35 , 0   ]

# define min and max for R values of Poincare map launch point
# Order given by:
# 'eim', 'ejm252', 'ejm004', 'flm001',
# 'fmm', 'fmm001', 'fmm002', 'fmm003', 
# 'fnm', 'fnm001', 'fnm002', 'fom003', 'fqm001'
Poincare_Rmin= {'eim'    : 6.17,
                'eim_fix' : 6.17,
                'ejm252' : 6.17,
                'ejm004' : 6.17,
                'flm001' : 6.12, 
                'fmm'    : 6.12,
                'fmm001' : 6.12,
                'fmm002' : 6.12,
                'fmm003' : 6.12,
                'fnm'    : 6.12,
                'fnm001' : 6.12,
                'fnm002' : 6.12,
                'fom003' : 6.11,
                'fqm001' : 6.07 }
Poincare_Rmax= {'eim'    : 6.28,
                'eim_fix' : 6.28,
                'ejm252' : 6.28,
                'ejm004' : 6.28,
                'flm001' : 6.23, 
                'fmm'    : 6.23,
                'fmm001' : 6.23,
                'fmm002' : 6.23,
                'fmm003' : 6.23,
                'fnm'    : 6.23,
                'fnm001' : 6.23,
                'fnm002' : 6.23,
                'fom003' : 6.22,
                'fqm001' : 6.18 }


def make_plots(this_config):
    print('<----Starting to work on ', this_config)
    # Step 2
    # Make R,Z, phi values
    # define several mesh grids - they all use the different Rmin, Rmax, Zmin, Zmax 
    # and have different Phi coordinates 

    reff_inversion_array = {}
#    reff_inversion_array_bulk = {}
    rmesh_plot = {}
    zmesh_plot = {}
    minor_radius = {}

 #   this_range = [0,4, 6, 2, 8]
    this_range = [3, 7, 1, 9, 5]

 #   this_range = [3]
    # For each toroidal cut
    # for ii in range(0,10):
    for ii in this_range:# range(0,10):
        print('<----Constructing R, Phi, Z')
        # Make R,Z, phi values

        Rline = np.linspace(Rmin[ii], Rmax[ii], num_pts_grid)
        Zline = np.linspace(Zmin[ii], Zmax[ii], num_pts_grid)
        Rmesh_temp, Zmesh_temp = np.meshgrid(Rline, Zline)

        Rmesh = Rmesh_temp
        Phimesh =  0 * Rmesh  + tor_angles[ii]
        Zmesh = Zmesh_temp

        rmesh_plot[ii] = Rmesh
        zmesh_plot[ii] = Zmesh        
    
        # Get 'rho', 'r_eff', for each cut
        print('<----Calling inversion for toroidal plane #', str(ii))
        tic = time.time()
        minor_radius[ii] = 0
        s_inversion, reff_inversion = invert_coords(vmec_configs[this_config], Rmesh, Phimesh, Zmesh, inv_tol)

        reff_inversion_array[ii] = np.array(s_inversion)

        toc = time.time()
        print('<----Completed inversion in ' +str(toc-tic) + ' seconds')

        


    #Step 3
    # Do field line following for configuration, starting at 'reasonable' locations
    # Generate Poincare Map for toroidal cut of interest

    pos.x1 = np.linspace(Poincare_Rmin[this_config], Poincare_Rmax[this_config], num_poincare_launch_points)
    pos.x2 = np.zeros(num_poincare_launch_points)
    pos.x3 = np.zeros(num_poincare_launch_points)

    poincare = fieldline_client.types.PoincareInPhiPlane()
    poincare.numPoints = num_poincare_points_per_plane

    poincare.phi0 = tor_angles
    
    task = fieldline_client.types.Task()
    task.step = 0.01
    task.poincare = poincare
    
    PoincareMaps = {}

    print('<----Starting Poincare map for ' + this_config)
    tic = time.time()
    PoincareMaps = fieldline_client.service.trace(pos, 
                                linefollow_configs[this_config],
                                task, None, None)
    toc = time.time()
    print('<----Completed Poincare map in ' +str(toc-tic) + ' seconds')
    
    
    #Step 4: Combine data from 1 and 2 into a single (set of) figures(s)

    
    
    if True:
        plt.figure()
        for ii in this_range:
    
            reff_inversion_plot = reff_inversion_array[ii]

            ind_gt1 = (reff_inversion_plot > 1.00).nonzero()
            for [thisrow,thiscol] in zip(ind_gt1[0], ind_gt1[1]):
                reff_inversion_plot[thisrow, thiscol] = np.NaN

            print('<----Attempting to save data ii=' + str(ii))
            save_filename = 'contour_data_' + this_config + '_hp_' + str(ii) + '.hdf5'
            outfile_fullpath = os.path.join(save_filename)
            outfile = h5py.File(outfile_fullpath, "w")
            dataset = outfile.create_dataset("rmesh_plot",
                                             data=rmesh_plot[ii])
            dataset = outfile.create_dataset("zmesh_plot",
                                        data=zmesh_plot[ii])
            dataset = outfile.create_dataset("rho_inversion",
                                        data=np.sqrt(reff_inversion_plot))

            outfile.close()
            print('<----Done saving data')
                
                
            plt.contourf(rmesh_plot[ii], zmesh_plot[ii], np.sqrt(reff_inversion_plot), cmap='jet', levels=[0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0])  #, vmin=0, vmax=.15)

        for kk in range(0,num_poincare_launch_points):
            if (np.mod(ii,2) == 0):
                jj_start = 0
                myaxes = [5.1, 6.3, -1.2, 1.2]
            else:
                jj_start = 1
                myaxes = [4.1, 6.8, -0.8, 0.8]
                
            for jj in range(jj_start, num_toroidal_angles, 2):
                try:
                    this_index = (kk)*num_toroidal_angles + jj
    
                    this_surf = PoincareMaps.surfs[this_index]
        
                    phi_data = this_surf.phi0
                    #print('<---phi_data = ', str(phi_data))
                    x_data = this_surf.points.x1
                    y_data = this_surf.points.x2
                    z_data = this_surf.points.x3        
        
                    x = np.array(x_data)
                    y = np.array(y_data)
                    z = np.array(z_data)
                    R = np.sqrt(x**2 + y**2)

                    print('<----Attempting to save data rad,hp=' + str(kk) +
                           ',' + str(jj))
                    save_filename = ('poincare_data_' + this_config + '_rad_' +
                                        str(kk) + '_hp_' + str(jj) + '.hdf5')
                    outfile_fullpath = os.path.join(save_filename)
                    outfile = h5py.File(outfile_fullpath, "w")
                    dataset = outfile.create_dataset("R",
                                                     data=R)
                    dataset = outfile.create_dataset("Z",
                                                data=z)
    
                    outfile.close()
                    print('<----Done saving data')
                



                    plt.hold(True)
                    plt.plot(R, z, ".k", ms = 1.0) #zeigt nur jede zweite Ebene an

                    plt.xlabel(r"$R$ [m]", fontsize=16)
                    plt.ylabel(r"$Z$ [m]", fontsize=16)
                    plt.grid()
                except:
                    print('<----failure on ii=' + str(ii) + ' jj=' + str(jj) + ' kk=' + str(kk))
        
        plt.xlabel(r"$R$ [m]", fontsize=16)
        plt.ylabel(r"$Z$ [m]", fontsize=16)
        plt.grid()
        plt.axis(myaxes)
        
        
        title_str = str.upper(this_config)
        plt.title(title_str)
            
        mycb = plt.colorbar()
        mycb.set_label('Effective Radius')
        plt.axis([4.55, 6.2, -0.6, 0.6])
        plt.show()
        print('<---Showing figure')
    
    
    print('<----Finished!')


def invert_coords(vmec_config, R, Phi, Z, inv_tol):
        # getReff and toVMECCoordinates don't work with 2-d meshes. Convert them to
        # 1-d arrays
        Rm_1d = R.reshape(-1)
        Phim_1d = Phi.reshape(-1)
        Zm_1d = Z.reshape(-1)
        
        vp = vmec_client.types.Points3D()
        vp2 = vmec_client.types.Points3D()
        
        # X Y Z
        vp.x1 = Rm_1d * np.cos(Phim_1d)
        vp.x2 = Rm_1d * np.sin(Phim_1d)
        vp.x3 = Zm_1d
        # R Phi Z
        vp2.x1 = Rm_1d
        vp2.x2 = Phim_1d
        vp2.x3 = Zm_1d


        vmec_coords_vp_1d = vmec_client.service.toVmecCoordinatesExtrapolated(vmec_config, vp2, inv_tol)
        #vmec_coords_reff_1d = vmec_client.service.getReff(vmec_config, vp)
        s_vp_1d = np.array(vmec_coords_vp_1d.x1)
        #r_1d = np.array(vmec_coords_reff_1d)
        
        s_vp = np.reshape(s_vp_1d, (-1, num_pts_grid))
        #r = np.reshape(r_1d, (-1, num_pts_grid))
        return s_vp, np.nan*s_vp



#
#
# Unccoment the configurations you want to inspect
#
#
   
#make_plots('eim')
##make_plots('eim_fix')
make_plots('ejm252')       #
make_plots('ejm004')       #
make_plots('flm001')       #
make_plots('fmm')          #
make_plots('fmm001')       #
##make_plots('fmm002')  
##make_plots('fmm003')
make_plots('fnm')          #
make_plots('fnm001')       #
make_plots('fnm002')       #
##make_plots('fom003')
make_plots('fqm001')       #


# Or, run all of the the cases!

#for case_iter in all_cases:
#    print('<----Checking ', case_iter)
#    make_plots(case_iter)
